/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Performances           *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* AlsaMidiIn.hpp
   Gabriel M. Beddingfield
   2008-02-08

*/

#ifndef NDEBUG
#include <iostream>
#include <iomanip>
#endif

#include <alsa/asoundlib.h>
#include <alsa/seq.h>

#include "AlsaMidiIn.hpp"
#include "TimeBase.hpp"
#include "MidiImplementation.hpp"

using namespace InConcert;

AlsaMidiIn::AlsaMidiIn(TimeBase* tb) :
    m_timebase(tb),
    m_seq(0),
    m_input(0)
{
    snd_seq_open(&m_seq, "default", SND_SEQ_OPEN_INPUT, 0);
    m_input = snd_seq_create_simple_port(
        m_seq,
        "InConcert MIDI In",
        SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
        SND_SEQ_PORT_TYPE_MIDI_GENERIC|SND_SEQ_PORT_TYPE_APPLICATION
        |SND_SEQ_PORT_TYPE_SOFTWARE);
}

AlsaMidiIn::~AlsaMidiIn()
{
    snd_seq_delete_simple_port(m_seq, m_input);
}


void AlsaMidiIn::run(void)
{
    snd_seq_event_t* ev;
    unsigned char buf[16];
    size_t len;
    int rv;
    do {
        rv = snd_seq_event_input(m_seq, &ev);
        if( rv == -EAGAIN ) {
            sleep(1);
            continue;
        } else if ( rv < 0 ) {
            break;
        }
        // Only process control events
        if(ev->type == SND_SEQ_EVENT_CONTROLLER) {
            buf[0] = 0xB0 | ev->data.control.channel;
            buf[1] = ev->data.control.param;
            buf[2] = ev->data.control.value;
            len = 3;
            ProcessMidiEvent(buf, len);
        } else {
#ifndef NDEBUG
            std::cerr << "Event filtered out" << std::endl;
#endif
        }
    } while (rv >= 0);
}

int AlsaMidiIn::ProcessMidiEvent(const unsigned char* buf, const size_t len)
{
    if(len == 0) return 0;
#ifndef NDEBUG
    for(size_t k=0; k<len; ++k) {
        std::cerr << std::hex << ((unsigned int)buf[k]) << " ";
    }
    std::cerr << std::endl;
#endif
    if( (len > 2)   // CC's only
        && (0xB0 == (buf[0] & 0xF0))
        && (buf[1] >= MIDI_CC_MIN)
        && (buf[1] <= MIDI_CC_MAX) ) {

        // Filter events
        if(buf[1] == CC_SET_CHANNEL) {
#ifndef NDEBUG
            std::cerr << "CC_SET_CHANNEL" << std::endl;
#endif
            if(buf[2] < 0x0F) {
                m_chan = buf[2];
#ifndef NDEBUG
                std::cerr << "Listening on channel " << std::hex << ((unsigned)m_chan) << std::endl;
#endif
            }
            return 0;
        }
        assert(m_chan <= 0x0F);
        if((buf[0] & 0x0F) != m_chan) {
#ifndef NDEBUG
            std::cerr << "Ignoring." << std::endl;
#endif
            return 0;
        }

        // Process the events that get through.
        switch(buf[1]) {
        case CC_REZERO:
#ifndef NDEBUG
            std::cerr << "CC_REZERO" << std::endl;
#endif
            m_timebase->re_zero();
            break;
        case CC_TAP:
#ifndef NDEBUG
            std::cerr << "CC_TAP" << std::endl;
#endif
            if(buf[2] & 0x40) m_timebase->tap();
            break;
        case CC_STARTSTOP:
#ifndef NDEBUG
            std::cerr << "CC_STARTSTOP" << std::endl;
#endif
            if(buf[2] & 0x40) {
                m_timebase->play();
            } else {
                m_timebase->stop();
            }
            break;
        case CC_INSERTORSKIP:
#ifndef NDEBUG
            std::cerr << "CC_INSERTORSKIP" << std::endl;
#endif
            if(buf[2] & 0x40) {
                m_timebase->skip_beat();
            } else {
                m_timebase->insert_beat();
            }
            break;
        case CC_TAPSREQUIRED:
#ifndef NDEBUG
            std::cerr << "CC_TAPSREQUIRED" << std::endl;
#endif
            if((buf[2] > 2)
               && (buf[2] <= TAPS_MAX)) {
                m_timebase->setTapsRequired((unsigned)buf[2]);
#ifndef NDEBUG
                std::cerr << "Now requiring " << m_timebase->getTapsRequired() << " taps" << std::endl;
#endif
            }
            break;
        case CC_TAPSCUECTL:
#ifndef NDEBUG
            std::cerr << "CC_TAPSCUECTL" << std::endl;
#endif
            if(buf[2] & 0x40) {
                m_timebase->setAutoStart(true);
            } else {
                m_timebase->setAutoStart(false);
            }
            break;
        case CC_BEATSPERMEAS:
#ifndef NDEBUG
            std::cerr << "CC_BEATSPERMEAS" << std::endl;
#endif
            if(buf[2]>0) m_timebase->setBeatsPerMeasure((unsigned)buf[2]);
            break;
        case CC_SETTEMPO:
#ifndef NDEBUG
            std::cerr << "CC_SETTEMPO" << std::endl;
#endif
            if(unsigned(buf[2])<= 127) m_timebase->setTempo(25.5 + double(buf[2])*1.50);
            break;
        default:
            assert(false);  // How did we get here... this event should have been filtered
        }
    }
    return 0;
}
