/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* Epoch.cpp
   Gabriel M. Beddingfield
   2008-08-29

*/

#include <cmath>
#include <cassert>

#include <jack/transport.h>
#include <jack/types.h>

#include "Epoch.hpp"

using namespace std;
using namespace InConcert;

/**
 * Initialize the internal sample-rate-based values
 *
 * @param[in] rate the number of audio frames (samples) per second;
 * the sampling rate.
 */
void Epoch::init_frame_rates(jack_nframes_t rate) {
    m_anchor_beat_abs = (double)(m_bar-1.0)*m_beats_per_bar
        + (m_beat-1.0)
        + (double)m_tick/m_ticks_per_beat;
    m_frames_per_beat = (double)rate * 60.0 / m_beats_per_minute;
    m_frames_per_tick = m_frames_per_beat / m_ticks_per_beat;
    m_frames_per_measure = m_frames_per_beat * m_beats_per_bar;
}

/**
 * Fill out the values in a jack_position_t struct to match the Epoch's position
 *
 * This function assumes that both the Epoch struct and the
 * jack_position_t struct are properly initialized and describing the
 * same song/piece.
 *
 * @param[out] pos the jack_position_t struct that describes the
 * current position within the song.
 */
void Epoch::set_jack_position_meta(jack_position_t* pos) {
    // Thanks to Dominic Sacré <dominic.sacre@gmx.de> for suggesting this
    // workaround to cast an 'int' to 'jack_position_bits_t'.
    pos->valid = static_cast<jack_position_bits_t>(JackPositionBBT | JackBBTFrameOffset);
    pos->beats_per_bar = (int32_t)m_beats_per_bar;
    pos->beat_type = 4;
    pos->ticks_per_beat = (int32_t)m_ticks_per_beat;
    pos->beats_per_minute = m_beats_per_minute;
}

/**
 * Displays the values of the class as a string
 *
 * @param os an ostream object (e.g. std::cout)
 * @param x an Epoch object
 */
ostream& operator<<(ostream& os, Epoch x)
{
    os << "[" << x.m_bar
       << ":" << x.m_beat
       << ":" << x.m_tick
       << "] " << x.m_frame
       << " (" << x.m_anchor_beat_abs
       << ") @ " << x.m_frames_per_beat
       << " f/beat (" << x.m_beats_per_minute
       << " bpm) and |" << x.m_beats_per_bar
       << ":" << x.m_ticks_per_beat << "|";
    return os;
}

Epoch::Epoch(void)
{
}

/**
 * Copy constructor.
 *
 * This is required so that the mutex is not copied.
 *
 * @param o the Epoch object to copy from.
 */
Epoch::Epoch(const Epoch& o) :
    m_frame(o.m_frame),
    m_bar(o.m_bar),
    m_beat(o.m_beat),
    m_tick(o.m_tick),
    m_beats_per_bar(o.m_beats_per_bar),
    m_ticks_per_beat(o.m_ticks_per_beat),
    m_beats_per_minute(o.m_beats_per_minute),
    m_frames_per_beat(o.m_frames_per_beat),
    m_frames_per_tick(o.m_frames_per_tick),
    m_frames_per_measure(o.m_frames_per_measure),
    m_anchor_beat_abs(o.m_anchor_beat_abs),
    m_mutex()
{
}

/**
 * Assignment operator
 *
 * @param o the Epoch object to copy from
 *
 * @return this class instance
 */
Epoch& Epoch::operator=(const Epoch& o)
{
    // Locking both mutexes for write is used to avoid deadlocking
    // when we lock the two mutexes... and keep syntax tidy.  If you
    // know a better way, let me know.
    //
    // See:
    // http://docs.huihoo.com/boost/1-33-1/doc/html/threads/faq.html#id2786744
    // http://www.open-std.org/jtc1/sc22/wg21/docs/papers/2007/n2320.html#LockAlgoExamples
    // http://archives.free.net.ph/message/20050908.034338.8e7b1281.en.html

    if (this != &o) {
        QWriteLocker a((&m_mutex < &o.m_mutex) ? &m_mutex : &o.m_mutex);
        QWriteLocker b((&m_mutex > &o.m_mutex) ? &m_mutex : &o.m_mutex);

        m_frame = o.m_frame;
        m_bar = o.m_bar;
        m_beat = o.m_beat;
        m_tick = o.m_tick;
        m_beats_per_bar = o.m_beats_per_bar;
        m_ticks_per_beat = o.m_ticks_per_beat;
        m_beats_per_minute = o.m_beats_per_minute;
        m_frames_per_beat = o.m_frames_per_beat;
        m_frames_per_tick = o.m_frames_per_tick;
        m_frames_per_measure = o.m_frames_per_measure;
        m_anchor_beat_abs = o.m_anchor_beat_abs;
    }

    return *this;
}

Epoch::~Epoch()
{
}

/**
 * Calculates the bar:beat:tick of the given frame, based on the
 * properties of the Epoch struct.
 *
 * Note that the bar, beat, and tick returned may also be pointers
 * to this class's bar, beat, and tick.
 *
 * @param[in] frame the audio sample for which to determine the BBT.
 *
 * @param[out] bar pointer to return the bar/measure. Must be
 * non-NULL. The pointer may point to this->bar.
 *
 * @param[out] beat pointer to return the beat within the
 * measure. Must be non-NULL. The pointer may point to this->beat.
 *
 * @param[out] tick pointer to return the tick within the beat. Must
 * be non-NULL. The pointer may point to to this->tick.
 *
 * @param[out] frames pointer to return the remainder of audio frames
 * (samples) after the tick. Must be non-NULL.
 */
void Epoch::getBBT(jack_nframes_t frame,
                   int32_t* _bar,
                   int32_t* _beat,
                   int32_t* _tick,
                   jack_nframes_t* _frames) const
{
    /* IMPLEMENTATION NOTE:
     *
     * Don't forget that _bar could be this->bar, etc.
     */
    double dbeat, bar, beat, tick, frames;
    QReadLocker mutex(&m_mutex);
    // Find the difference in beats from the anchor
    dbeat = ((double)frame - (double)m_frame)/m_frames_per_beat;
    // Now add in the anchor
    dbeat += m_anchor_beat_abs;
    assert(dbeat >= 0.0);
    // Now convert to bbt.
    tick = modf(dbeat, &beat) * m_ticks_per_beat;
    mutex.unlock();  // We're done with the data
    bar = beat/m_beats_per_bar; // floor later
    beat = fmod(beat, m_beats_per_bar);
    frames = modf(tick, &tick) * m_frames_per_tick;

    assert(bar >= 0.0);
    assert(beat >= 0.0);
    assert(tick >= 0.0);
    assert(frames >= 0.0);

    (*_bar) = static_cast<int32_t>(bar)+1;             // assuming typecast floors()
    (*_beat) = static_cast<int32_t>(round(beat+1.0));  // beat should already be nearly an integer
    (*_tick) = static_cast<int32_t>(tick);             // assuming typecast floors()
    (*_frames) = static_cast<int32_t>(frames);         // assuming floor()
}

/**
 * Calculates the frame number of the given bar:beat:tick, based on
 * the properties of the Epoch struct.
 *
 * @param[in] bar the bar/measure number within the song.
 * @param[in] beat the beat within the bar
 * @param[in] tick the tick within the beat
 * @param[in] frames the number of audio frames (samples) after the tick
 *
 * @return the number of audio frames (samples) since the beginning of
 * the song.
 */
jack_nframes_t Epoch::getFrame(double bar, double beat, double tick, double frames) const
{
    QReadLocker mutex(&m_mutex);
    // Calculate absolute beats (decimal w/ticks)
    double f = (bar-1.0)*m_beats_per_bar
        + (beat-1.0)
        + tick/m_ticks_per_beat
        + frames;
    // offset by the anchor
    f -= m_anchor_beat_abs;
    // Convert to frames
    f *= m_frames_per_beat;
    // and offset by anchor
    f += m_frame;
    mutex.unlock();

    if( f < 0.0 ) {
        return (jack_nframes_t) 0;
    }
    return (jack_nframes_t) round(f);
}
