/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version                                                               *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* Client.cpp
 * 2008-08-17
 */

#include <iostream>
#include <cassert>
#include "Client.hpp"
#include "Port.hpp"
#include "TimebaseMaster.hpp"
#include <jack/jack.h>

using namespace std;

/**
 * The High-Level JACK API
 *
 * This is an experimental API to provide a high-level, C++ API for
 * the JACK Audio Connection Kit (JACK). It is not completely clear if
 * this is even a worthwhile endeavor since the JACK API is fairly
 * simple and very flexable. The primary rationale for this API is
 * that it helps to manage the relationships between the jack_client_t
 * pointer and the misc. objects that are spawned from it
 * (e.g. ports).
 *
 * The central class is the Client, which acts as a sort of parent for
 * all the other classes.
 */
namespace HiJack
{

jack_client_t* Client::ref(void)
{
    return m_client;
}

Client::Client(const std::string& name)
    : m_client(0),
      m_name(name),
      m_timebase(0)
{
    QMutexLocker mx(&m_mutex);
    open_locked();
    activate_locked();
}

#define CLIENT_FAILURE(msg) {                                           \
        cerr << "Could not connect to JACK server (" << (msg) << ")"    \
             << endl;                                                   \
        if (m_client) {                                                 \
            cerr << "...but JACK returned a non-null pointer?" << endl; \
            (m_client) = 0;                                             \
        }                                                               \
        if (tries) cerr << "...trying again." << endl;                  \
    }


#define CLIENT_SUCCESS(msg) {                   \
        assert(m_client);                       \
        cerr << msg << endl;                    \
        tries = 0;                              \
    }

void Client::open_locked(void)
{

    jack_status_t status;
    int tries = 2;  // Sometimes jackd doesn't stop and start fast enough.
    while ( tries > 0 ) {
        --tries;
        m_client = jack_client_open(
            m_name.c_str(),
            JackNullOption,
            &status);
        switch(status) {
        case JackFailure:
            CLIENT_FAILURE("unknown error");
            break;
        case JackInvalidOption:
            CLIENT_FAILURE("invalid option");
            break;
        case JackNameNotUnique:
            if (m_client) {
                m_name = jack_get_client_name(m_client);
                CLIENT_SUCCESS(string("Jack assigned the client name '")
                               + m_name + "'");
            } else {
                CLIENT_FAILURE("name not unique");
            }
            break;
        case JackServerStarted:
            CLIENT_SUCCESS("JACK Server started for Hydrogen.");
            break;
        case JackServerFailed:
            CLIENT_FAILURE("unable to connect");
            break;
        case JackServerError:
            CLIENT_FAILURE("communication error");
            break;
        case JackNoSuchClient:
            CLIENT_FAILURE("unknown client type");
            break;
        case JackLoadFailure:
            CLIENT_FAILURE("can't load internal client");
            break;
        case JackInitFailure:
            CLIENT_FAILURE("can't initialize client");
            break;
        case JackShmFailure:
            CLIENT_FAILURE("unable to access shared memory");
            break;
        case JackVersionError:
            CLIENT_FAILURE("client/server protocol version mismatch");
        default:
            if (status) {
                cerr << "Unknown status with JACK server." << endl;
                if (m_client) {
                    cerr << "Client pointer is *not* null..."
                        " assuming we're OK" << endl;
                }
            } else {
                CLIENT_SUCCESS("Connected to JACK server");
            }
        }
    }

    // Here, m_client should either be valid, or NULL.
}

Client::~Client()
{
    close_locked();
}

void Client::close_locked(void)
{
    int rv;
    if(m_client) {
        deactivate_locked();
        port_list_t::iterator pos;
        for( pos = m_ports.begin() ; pos != m_ports.end() ; ++pos ) {
            (*pos)->jackIsClosing();
        }
        if (m_timebase) {
            m_timebase->jackIsClosing();
        }
        rv = jack_client_close(m_client);  // Ignore return value
        if (rv) cerr << "jack_client_close(m_client) reported an error" << endl;
        m_client = 0;
    }
}

void Client::activate_locked(void)
{
    int rv = jack_activate(m_client);
    if (rv) cerr << "jack_activate(m_client) reported and error" << endl;
}

void Client::deactivate_locked(void)
{
    int rv = jack_deactivate(m_client);
    if (rv) cerr << "jack_deactive(m_client) reported an error" << endl;
}

const std::string& Client::getName(void) const
{
    return m_name;
}

bool Client::registerPort(Port* port)
{
    QMutexLocker mx(&m_mutex);
    jack_port_t* j_port;
    const char* type;
    bool rv = false;
    if (port->getType() == Port::Midi) {
        type = JACK_DEFAULT_MIDI_TYPE;
    } else {
        type = JACK_DEFAULT_AUDIO_TYPE;
    }
    j_port = jack_port_register(m_client,
                                port->getName().c_str(),
                                type,
                                port->getFlags(),
                                0);
    if (j_port) {
        port->m_port = j_port;
        port->m_client = this;
        pair<port_list_t::iterator, bool> rv_i;
        rv_i = m_ports.insert(port);
        rv = rv_i.second;
    } else {
        cerr << "Could not register port " << port->getName()
             << "with JACK." << endl;
    }
    return rv;
}

bool Client::unregisterPort(Port* port)
{
    bool rv = true;
    int jerr = 0;
    QMutexLocker mx(&m_mutex);
    deactivate_locked();
    jerr = jack_port_unregister(m_client, port->m_port);
    if (jerr) {
        cerr << "JACK gave an error when unregistering " << port->getName()
             << endl;
        rv = false;
    }
    removePortFromProcessCallback_locked(port);
    port->m_client = 0;
    int erased = m_ports.erase(port);
    assert(erased > 0);
    activate_locked();
    return rv;
}

int Client::jackProcessCallback(jack_nframes_t nframes, void* arg)
{
    Client* me = static_cast<Client*>(arg);
    if (me == 0) {
        return -1;
    }
    QMutexLocker mx(&(me->m_mutex));
    port_list_t::iterator pos;
    int rv = 0;
    for( pos = me->m_process.begin() ; pos != me->m_process.end() ; ++pos ) {
        rv += (*pos)->processCallback(nframes);
    }
    return rv;
}

// Returns true if port on the left should be called earlier in
// the process callback than the port on the right.
// Order:
// Midi Input (3), Audio Input (2), Midi Output (1), Audio Output (0)
// (Note, this makes bit 0 (Midi|Audio), and bit 1 (Input|Output)).
// If any of these is a tie, the tie is broken by the value of
// the pointer to the port.
bool Client::procSort::operator()(Port* left, Port* right)
{
    int lscore, rscore;

    lscore = ((left->getType() == Port::Midi) ? 1 : 0)
        + ((left->getFlags() & JackPortIsInput) ? 2 : 0);
    rscore = ((right->getType() == Port::Midi) ? 1 : 0)
        + ((right->getFlags() & JackPortIsInput) ? 2 : 0);

    if (lscore == rscore) {
        return left < right;
    }
    return (lscore > rscore);
}

bool Client::addPortToProcessCallback(Port* port) {
    bool rv = false;
    QMutexLocker mx(&m_mutex);
    bool jreg = (m_process.size() == 0);
    if (m_client && (m_process.find(port) != m_process.end())) {
        m_process.insert(port);
        rv = true;
    }
    if (jreg) {
        deactivate_locked();
        int err = jack_set_process_callback(m_client,
                                            jackProcessCallback,
                                            this);
        if (err) cerr << "JACK rejected the process callback..."
                      << " this is unrecoverable" << endl;
        activate_locked();
    }
    return rv;
}

bool Client::removePortFromProcessCallback(Port* port) {
    QMutexLocker mx(&m_mutex);
    return removePortFromProcessCallback_locked(port);
}

bool Client::removePortFromProcessCallback_locked(Port* port) {
    bool rv = false;
    port_list_t::iterator pos = m_process.find(port);
    if (m_client && (pos != m_process.end())) {
        m_process.erase(pos);
        rv = true;
    }
    if (m_process.size() == 0) {
        jack_set_process_callback(m_client, 0, 0);
    }
    return rv;
}

bool Client::registerTimebaseMaster(TimebaseMaster* master, bool if_none_already)
{
    int rv;
    QMutexLocker mx(&m_mutex);
    if (master == 0) return false;
    if (m_timebase && if_none_already) return false;
    if (m_timebase) {
        m_timebase->m_client = 0;
        m_timebase = master;
        m_timebase->m_client = this;
        return true;
    }
    rv = jack_set_timebase_callback(m_client,
                                    if_none_already,
                                    jackTimebaseCallback,
                                    static_cast<void*>(this));
    if (rv == 0) {
        m_timebase = master;
        m_timebase->m_client = this;
        return true;
    } // else
    return false;
}

// If tbm is 0, then unregister whatever we currently have for a timebase.
bool Client::unregisterTimebaseMaster(TimebaseMaster* tbm)
{
    bool rv = false;
    int err = 0;
    QMutexLocker mx(&m_mutex);
    if (m_timebase && ((m_timebase == tbm) || (tbm == 0))) {
        err = jack_release_timebase(m_client);
        if (err == 0) {
            m_timebase->m_client = 0;
            m_timebase = 0;
            rv = true;
        }
    }
    return rv;
}

void Client::jackTimebaseCallback(jack_transport_state_t state,
                                  jack_nframes_t nframes,
                                  jack_position_t *pos,
                                  int new_pos,
                                  void* arg)
{
    Client *me = static_cast<Client*>(arg);
    if (me && me->m_timebase) {
        me->m_timebase->timebaseCallback(state, nframes, pos, new_pos);
    }
}

TimebaseMaster* Client::getTimebaseMaster(void) const
{
    return m_timebase;
}

//     std::vector<QString> Client::getMidiOutputPortList(void)
//     {
//         vector<QString> ports;
//         const char **port_names = 0;
//         port_names = jack_get_ports(m_client,
//                                     0,
//                                     JACK_DEFAULT_MIDI_TYPE,
//                                     JackPortIsOutput);
//         if (!port_names) return ports;
//         int k = 0;
//         while (port_names[k]) {
//             ports.push_back(QString(port_names[k]));
//             ++k;
//         }
//         free((void*)port_names);
//         return ports;
//     }

} // namespace HiJack

