#ifndef INCONCERT_TAPTEMPOBASEIMPL_HPP__
#define INCONCERT_TAPTEMPOBASEIMPL_HPP__
/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* TapTempoBaseImpl.hpp
   Gabriel M. Beddingfield
   2008-08-16

 */

#include <string>
#include <cassert>
#include <stdexcept>
#include "TapTempoInterface.hpp"

namespace InConcert
{

/*
 * class TapTempoBaseImpl
 *
 * This class provides basic implementations that most
 * other TapTempo models need.
 *
 */
template <typename ScalarT, typename TempoT = double>
class TapTempoBaseImpl : public TapTempoInterface<ScalarT, TempoT>
{
    std::string m_name;
    unsigned m_taps_required;
    TempoT m_rate_multiplier;
    ScalarT m_timeout;
    ScalarT m_tempo_interval;
    TempoT m_tempo_rate;
    bool m_reset;  // reset the object the next time tap() is called.

public:
    TapTempoBaseImpl(const std::string& name_) :
        m_name(name_),
        m_taps_required(4),
        m_rate_multiplier((ScalarT)1),
        m_timeout(0),
        m_tempo_interval(0),
        m_tempo_rate(0),
        m_reset(false)
    {
    }

    virtual ~TapTempoBaseImpl()
    {
    }

    bool reset(void) { return m_reset; }
    void queueReset(void) { m_reset = true; }

    void setTempoByInterval(ScalarT interval)
    {
        assert(interval > 0);
        m_tempo_interval = interval;
        m_tempo_rate = m_rate_multiplier/TempoT(interval);
    }

    void setTempoByRate(TempoT rate)
    {
        assert(rate > 0);
        m_tempo_rate = rate;
        m_tempo_interval = m_rate_multiplier/rate;
        assert( m_tempo_interval > 0 );
    }

    /*
     * See doc for TapTempoInterface for documentation on these:
     */
    virtual const std::string& name(void) const { return m_name; }
    virtual unsigned getTapsRequired(void) const { return m_taps_required; }

    virtual void setTapsRequired(unsigned t)
    {
        if (t < 2) {
            throw std::out_of_range("TapTempoBaseImpl::setTapsRequired: t must be 2 or more");
        }
        m_taps_required = t;
    }

    virtual const TempoT& getRateMultiplier(void) const
    {
        return m_rate_multiplier;
    }

    virtual void setRateMultiplier(TempoT rate)
    {
        if (rate <= 0) {
            throw std::out_of_range("TapTempoBaseImpl::setRateMultiplier: rate must be greater than 0");
        }
        m_rate_multiplier = rate;
    }

    virtual const ScalarT& getTimeout(void) const
    {
        return m_timeout;
    }

    virtual void setTimeout(ScalarT t)
    {
        if (t <= 0) {
            throw std::out_of_range("TapTempoBaseImpl::setTimeout: t must be greater than 0");
        }
        m_timeout = t;
    }

    virtual void invalidate(void)
    {
        m_tempo_interval = 0;
        m_tempo_rate = 0;
        m_reset = false;
    }

    virtual ScalarT tap(ScalarT tap_) = 0;

    virtual const TempoT& getTempoRate(void) const { return m_tempo_rate; }
    virtual const ScalarT& getTempoInterval(void) const { return m_tempo_interval; }

}; // class TapTempoBaseImpl

} // namespace InConcert

#endif // INCONCERT_TAPTEMPOBASEIMPL_HPP__
