#ifndef INCONCERT_TAPTEMPOINTERFACE_HPP__
#define INCONCERT_TAPTEMPOINTERFACE_HPP__
/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* TapTempoInterface.hpp
   Gabriel M. Beddingfield
   2008-08-16

 */

#include <string>

namespace InConcert
{

/**
 * A portable interface to different tap tempo models
 *
 * Since tap tempo models have some common concepts, this class
 * template defines a common interface to make it easier to
 * prototype, test, and harness different approaches to tap
 * prediction.
 *
 * Thread-safety: Implementations are must meet the following
 * requirements for threading:
 *
 *     * All member functions must be reentrant, changing only
 *       the state of the class and no static or global members.
 *
 *     * It is strongly recommended that the implementation
 *       <em>not</em> use a mutex.
 *
 *     * If an external read-lock is used, accessing the const
 *       functions of the class shall be safe and well-formed from
 *       multiple threads (no race conditions).
 *
 *     * If an external mutex (write-lock) is used, accessing the
 *       const and non-const methods shall be safe and well-formed
 *       from a single thread.
 *
 * @param ScalarT the type to use for measuring time. This can be
 * an unsigned type (e.g. number of audio samples) or it can be
 * floating point type (e.g. seconds). Implementers should be
 * careful to gracefully handle overflow conditions, and be aware
 * that many algorithms will easily overflow this type.
 *
 * @param TempoT the type to use for reporting the concept of
 * tempo (i.e. taps per unit of time). This must be a
 * floating-point type.
 *
 */
template <typename ScalarT, typename TempoT = double>
class TapTempoInterface
{
public:
    virtual ~TapTempoInterface() { }

    typedef ScalarT scalar_type;    //< the "time" type (integer or float)
    typedef TempoT tempo_type;      //< the "tempo" type (float)

    /**
     * Returns a friendly name for the tap-tempo implementation
     *
     * Each tap-tempo implementation must define a
     * user-recognizeable name for the implementation. There are
     * no rules for the name, but the expectation is that it can
     * be easily mapped back to the name of the
     * class. (E.g. "Least Squares" or "Rolling Average").
     *
     * @return name of tap tempo algorithm.
     */
    virtual const std::string& name(void) const = 0;

    /**
     * Get the number of taps required to predict next beat
     *
     * This is generally used for start-up and tempo-change logic,
     * and is driven by the needs of the musician rather than the
     * algorithm.
     *
     * @return after invalidate() has been called, this is the
     * number of times that tap() that must be called before it
     * will predict the next beat. The value shall be greater than
     * 1.
     */
    virtual unsigned getTapsRequired(void) const = 0;

    /**
     * Set the number of taps required to predict next beat
     *
     * This is generally used for start-up and tempo-change logic,
     * and is driven by the needs of the musician rather than the
     * algorithm.
     *
     * All algorithms must work with t==2. If a larger value is
     * supplied by the user, then it the algorithm may utilize the
     * extra information for the prediction.
     *
     * @param t after invalidate() has been called, this is the
     * number of times that tap() must be called before predicting
     * a beat. The value must be greater than 1.
     *
     * @throw std::out_of_range if t is less than 2.
     */
    virtual void setTapsRequired(unsigned t) = 0;

    /**
     * Get the rate multiplier for reporting "tempo"
     *
     * See setRateMultipler().
     *
     * @return the value multiplied against the raw
     * taps-per-scaler_type when reporting tempo.
     */
    virtual const TempoT& getRateMultiplier(void) const = 0;

    /**
     * Set the rate multiplier for reporting "tempo"
     *
     * The tempo is taps-per-time (i.e. taps-per-scalar_type).
     * However, these units are rarely useful (e.g. taps per audio
     * sample, taps per second). This allows you to set a
     * multipler for reporting tempo to convert the value from
     * taps-per-audio-sample to something like taps-per-minute.
     *
     * For example, if the current sample rate is 44,100 Hz, then
     * 120 beats/min is a sample interval of 22,050.  Meanwhile,
     * there are 60 * 44,100 = 2,646,000 samples in a minute.  By
     * setting 2,646,000 as the rate multipler, it converts the
     * rate to beats/min. So if the taps are spaced 22,050 apart
     * and the rate multiplier is 2,646,000 then the tempo
     * returned will be 2,646,000 / 22,050 = 120.0
     *
     * @param rate the conversion factor for the raw tempo. The
     * value must be non-zero, and should be greater than zero.
     *
     * @throw std::out_of_range if rate is zero.
     */
    virtual void setRateMultiplier(TempoT rate) = 0;

    /**
     * Get the timeout period for the model.
     *
     * See setTimeout()
     *
     * @return the amount of time (in units of scalar_type) after
     * which the model will reset.
     */
    virtual const ScalarT& getTimeout(void) const = 0;

    /**
     * Set the timeout period for the model.
     *
     * The model shall disregard all data that is older than t
     * units of time. The expectaction is that the model will
     * automatically invalidate() after this much time of
     * inactivity. However, there is also a weak expectation that
     * an active filter will have all taps older than t removed
     * from the internal model.
     *
     * @param t the amount of time of inactivity before
     * auto-invalidate()-ing. Must be greater than zero.
     *
     * @throw std::out_of_range if t is zero.
     */
    virtual void setTimeout(ScalarT t) = 0;

    /**
     * Reset the current engine's taps.
     *
     * After calling this, all record of taps will be dropped and
     * the beat prediction will return to default values. Note
     * that this does not affect properties, just the values
     * received from tap(). (i.e. things like getTapsRequired() or
     * getTimeout() don't change.)
     */
    virtual void invalidate(void) = 0;

    /**
     * Register a tap to define a new tempo and downbeat sync.  Supply
     * a reference time 'tap_' (e.g. frame number).
     *
     * @return 0 if the model is not able to predict the next beat.
     * If the model is able to predict the next beat, then will return
     * a non-zero value.  If it returns non-zero, then the tempo may
     * be obtained by calling getTempo().
     */
    virtual ScalarT tap(ScalarT tap_) = 0;

    /**
     * Get the current tempo estimation (taps-per-time)
     *
     * If tap() returns non-zero, this function will return an
     * estimate of the current tempo. This is a convenience
     * function for getTempoInterval().
     *
     * @return estimate of current tempo in the user-defined
     * units, getTempoRate() = getRateMultipler() /
     * getTempoInterval(). If tap() or getTempoInterval() returns
     * 0, this function returns 0.
     */
    virtual const TempoT& getTempoRate(void) const = 0;

    /**
     * Get the current tempo estimation (time-per-tap)
     *
     * If tap() returns non-zero, ths function will return an
     * estimate of the current tempo. The value returned will be
     * the number of time units (scalar_type) between each
     * beat/tap.
     *
     * @return the amount of time between beats (taps), in units
     * of scalar_type. If tap() returns 0, this function returns
     * 0.
     */
    virtual const ScalarT& getTempoInterval(void) const = 0;
}; // class TapTempoInterface

} // namespace InConcert

#endif // INCONCERT_TAPTEMPOINTERFACE_HPP__
