/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* TimeBase.cpp
   Gabriel M. Beddingfield
   2008-01-29

   Implementation of TimeBase class.  See header for documentation.
*/

// Standard C++ libraries
#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <cmath>
#include <cassert>

// POSIX libraries
#include <unistd.h>  // for usleep()

// Jack libraries
#include <jack/jack.h>
#include <jack/transport.h>
#include <jack/types.h>

// Qt Libraries (QtCore only)
#include <QMutexLocker>
#include <QReadWriteLock>
#include <QReadLocker>
#include <QWriteLocker>

#include "TimeBase.hpp"
#include "TapTempo/TapTempoMedian.hpp"
#include "HiJack/TimebaseMaster.hpp"
#include "HiJack/Client.hpp"
#include "global.hpp"

using namespace std;
using namespace InConcert;

TimeBase::TimeBase(void) :
    m_state(Steady),
    m_state_mutex(),
    m_time_reset(false),
    m_auto_start(false),
    m_name("InConcert::TimeBase")
{
    // Initialize m_normal
    m_normal.m_frame = 0;
    m_normal.m_bar = 1;
    m_normal.m_beat = 1;
    m_normal.m_tick = 0;
    m_normal.m_beats_per_bar = 4;
    m_normal.m_ticks_per_beat = 360;
    m_normal.m_beats_per_minute = 120.0;
    m_ramp = m_normal;

    m_taps.reset( new TapTempoMedian<jack_nframes_t, double> );
    HiJack::Client& client = getJackClientInstance();
    m_taps->setRateMultiplier(60 * jack_get_sample_rate(client.ref()));
}

TimeBase::~TimeBase()
{
}

/* MUST BE CALLED WITH m_state_mutex ALREADY LOCKED */
void TimeBase::timebase_ramping(
    jack_transport_state_t /*state*/,
    jack_nframes_t /*nframes*/,
    jack_position_t *pos,
    int new_pos)
{
    assert(pos->frame >= m_ramp.m_frame);
    assert(pos->frame < m_normal_queued.m_frame);
    assert(m_state == Ramping);
    //assert(m_state_mutex->status() == Mutex::Locked);

    if(new_pos || m_time_reset) {
        m_ramp.init_frame_rates(pos->frame_rate);
        m_ramp.set_jack_position_meta(pos);
        m_taps->setRateMultiplier(60 * pos->frame_rate);
        m_time_reset = false;
    }
    m_ramp.getBBT(pos->frame, &pos->bar, &pos->beat, &pos->tick, &pos->bbt_offset);
}

/* MUST BE CALLED WITH m_state_mutex ALREADY LOCKED */
void TimeBase::timebase_normal(
    jack_transport_state_t /*state*/,
    jack_nframes_t /*nframes*/,
    jack_position_t *pos,
    int new_pos)
{
    // assert((m_state != Ramping) && (pos->frame > m_anchor_frame));
    //assert(m_state_mutex->status() == Mutex::Locked);

    if(new_pos || m_time_reset) {
        m_normal.init_frame_rates(pos->frame_rate);
        m_normal.set_jack_position_meta(pos);
        m_taps->setRateMultiplier(60 * pos->frame_rate);
        m_time_reset = false;
    }
    m_normal.getBBT(pos->frame, &pos->bar, &pos->beat, &pos->tick, &pos->bbt_offset);
}

// Jack Timebase callback.
//
// This determines if we're in normal mode (Steady), ramping mode
// (Ramping), or if a ramp has been queued to start (RampQueued).  It
// calls the correct method depending on the state.
//
void TimeBase::timebaseCallback(
    jack_transport_state_t state,
    jack_nframes_t nframes,
    jack_position_t *pos,
    int new_pos)
{
    HiJack::Client& client = getJackClientInstance();
    QMutexLocker lock(&m_state_mutex);

    if(((m_state == Ramping)||(m_state == RampQueued))
       && (pos->frame >= m_ramp.m_frame)) {

        if(pos->frame < m_normal_queued.m_frame) {
            if(m_state == RampQueued) {
                m_state = Ramping;
                m_time_reset = true;
            }
            timebase_ramping(state, nframes, pos, new_pos);
        } else {
            cerr << "m_normal: " << m_normal << endl;
            cerr << "queued:   " << m_normal_queued << endl;
            m_normal = m_normal_queued;
            cerr << "Going steady: (nframes, rs, af, jack_frame_time) = ("
                 << pos->frame << ", "
                 << m_ramp.m_frame << ", "
                 << m_normal.m_frame << ", "
                 << jack_frame_time(client.ref())
                 << ")" << endl;
            m_state = Steady;
            m_time_reset = true;
            timebase_normal(state, nframes, pos, new_pos);
        }
    } else {
        timebase_normal(state, nframes, pos, new_pos);
    }
}

// TapTempo has given us a frame to snap our beat to.  The trick
// to making this function work is to find the nearest beat to
// snap to, and to make sure that we don't skip any frames by
// accidentally setting m_ramp in the past.
#warning "This has a hitch somewhere.  It drags the beat.  Feels like"
#warning "an 8th note or a 16th note at 120 bpm."
/* MUST BE CALLED WITH m_state_mutex ALREADY LOCKED */
void TimeBase::rampAndSnapTo(jack_nframes_t target, double new_tempo)
{
    Epoch ramp;
    Epoch normal = m_normal;
    HiJack::Client& client = getJackClientInstance();
    assert(JackTransportRolling == jack_transport_query(client.ref(), 0));
    jack_nframes_t frame_rate = jack_get_sample_rate(client.ref());
    jack_nframes_t ignored;

    // Find the nearest BBT for our new normal anchor
    normal.getBBT(target, &normal.m_bar, &normal.m_beat, &normal.m_tick, &ignored);
    normal.m_frame = target;
    if((normal.m_ticks_per_beat-normal.m_tick) <= normal.m_tick) {
        ++normal.m_beat;
    }
    normal.m_tick = 0;
    normal.m_beats_per_minute = round(10.0*new_tempo)/10.0;
    normal.init_frame_rates(frame_rate);

    // Set up the ramp.  Don't start the ramp until the current
    // frame + max_buffer_size
    ramp = m_normal;
    ramp.m_frame = jack_get_buffer_size(client.ref()) + jack_get_current_transport_frame(client.ref());
    assert(normal.m_frame > ramp.m_frame);
    m_normal.getBBT(ramp.m_frame, &ramp.m_bar, &ramp.m_beat, &ramp.m_tick, &ignored);
    ramp.init_frame_rates(frame_rate);
    assert(normal.m_anchor_beat_abs > ramp.m_anchor_beat_abs);
    ramp.m_beats_per_minute = 60.0*frame_rate*(normal.m_anchor_beat_abs - ramp.m_anchor_beat_abs)
        / (normal.m_frame - ramp.m_frame);
    ramp.init_frame_rates(frame_rate);

    cerr << "Old normal: " << m_normal << endl;
    cerr << "New normal: " << m_normal_queued << endl;
    cerr << "Ramp:       " << ramp << endl;
    m_normal_queued = normal;
    m_ramp = ramp;
    m_state = RampQueued;
}

void TimeBase::tap(void)
{
    tap(jack_frame_time( getJackClientInstance().ref() ));
}

// Register a tap for redefining the beat and tempo.
void TimeBase::tap(jack_nframes_t time)
{
    double new_tempo;
    jack_nframes_t target;
    HiJack::Client& client = getJackClientInstance();
    QMutexLocker lock(&m_state_mutex);

    try {
        target = m_taps->tap(time);
        cerr << "Now = " << time << " Target = " << target << endl;
        if(target) {
            new_tempo = m_taps->getTempoRate();
            if(JackTransportRolling == jack_transport_query(client.ref(), 0)) {
                // Convert the target to transport frames (estimate).
                target = target - (jack_frame_time(client.ref())
                                   - jack_get_current_transport_frame(client.ref()));
                rampAndSnapTo(target, new_tempo);
            } else {
                m_normal.m_beats_per_minute = new_tempo;
                m_time_reset = true;
                m_auto_start = 1;
                if(m_auto_start) {
#               if(_XOPEN_SOURCE == 500)
                    useconds_t usecs;
#               else
                    unsigned long usecs;
#               endif
                    jack_nframes_t jft = jack_frame_time(client.ref());
                    cerr << "target = " << dec << target << " frame_time = " << dec << jft << endl;
#warning "Hardcoded 10*buffer_size... need to get this from the output port."
#warning "...I think the *only* way is to do a config param or query ports."
                    // Estimate the number of microseconds from now that jack_frame_time()
                    // will hit frame #target.
                    // secs ~= (target-jft-latency)/samplerate
                    // usecs = secs * 1000000
                    usecs = floor(1000000.0*
                                  double(target-jft-10*jack_get_buffer_size(client.ref()))
                                  /double(jack_get_sample_rate(client.ref())));
                    cout << "usecs = " << dec << usecs;
                    double totsleep = usecs;
                    // usleep() over 1000000 usecs is not portable.
                    if (usecs >= 1000000) {
                        cerr << "macro usecs" << endl;
                        unsigned secs = usecs / 1000000;
                        usecs = usecs % 1000000;
                        #warning "Fast tapping causes a hang here.  secs == 2904, typ."
                        sleep(secs);
                        if(usecs) usleep(usecs);
                    } else {
                        usleep(usecs);
                    }
                    jack_transport_start(client.ref());
                    cerr << "Slept for " << dec << totsleep << " for tempo " << new_tempo
                         << " bpm (" << (60.0/new_tempo) << " sec/beat)" << endl;
                    cerr << "I estimated latency at " << jack_get_buffer_size(client.ref())*10 <<endl;
                    cerr << "Jack reports port latencies as:" << endl;
                    #warning "Leaking memory here."
                    const char** ports = jack_get_ports(client.ref(), 0, 0, 0);
                    jack_port_t* port;
                    while (*ports) {
                        port = jack_port_by_name(client.ref(), *ports);
                        cerr << "    " << (*ports) << ": "
                             << "jack_port_get_latency()="
                             << jack_port_get_latency(port)
                             << "  jack_port_get_total_latency() = "
                             << jack_port_get_total_latency(client.ref(), port)
                             << endl;
                        ++ports;
                    }
                }
            }
        }
    } catch (...) {

    }
    m_state_mutex.unlock();
}

// skip a beat by double-timing for one beat.
// Snaps to the downbeats.
//
// Create a new m_normal and m_ramp and queues the ramp (RampQueued).
//
void TimeBase::skip_beat(void)
{
    HiJack::Client& client = getJackClientInstance();

    if(m_state == Ramping) {
        cout << "Already ramping... not skipping" << endl;
        return;
    }
    cout << "skipping beat..." << endl;
    jack_position_t pos;
    jack_transport_query(client.ref(), &pos);
    assert(pos.valid == JackPositionBBT);

    Epoch ramp, normal;
    ramp.m_beats_per_bar = (int32_t)pos.beats_per_bar;
    ramp.m_ticks_per_beat = (int32_t)pos.ticks_per_beat;
    if(pos.beat == ramp.m_beats_per_bar) {
        ramp.m_bar = pos.bar + 1;
        ramp.m_beat = 1;
    } else {
        ramp.m_bar = pos.bar;
        ramp.m_beat = pos.beat + 1;
    }
    ramp.m_tick = 0;

    normal = ramp;
    if(normal.m_beat == normal.m_beats_per_bar) {
        ++normal.m_bar;
        normal.m_beat = 1;
    } else {
        ++normal.m_beat;
    }

    QMutexLocker lock(&m_state_mutex);
    ramp.m_frame = m_normal.getFrame(ramp.m_bar, ramp.m_beat, ramp.m_tick);
    normal.m_frame = m_normal.getFrame(normal.m_bar, normal.m_beat, normal.m_tick);
    ramp.m_beats_per_minute = 2.0 * pos.beats_per_minute;
    normal.m_beats_per_minute = pos.beats_per_minute;
    if(normal.m_beat == normal.m_beats_per_bar) {
        ++normal.m_bar;
        normal.m_beat = 1;
    } else {
        ++normal.m_beat;
    }
    normal.init_frame_rates(pos.frame_rate);
    ramp.init_frame_rates(pos.frame_rate);
    m_normal_queued = normal;
    m_ramp = ramp;
    m_state = RampQueued;

    // Ok, we set the ramp... now it's up to to
    // timebase() to do the Right Thing.
}

// skip a beat by half-timing for one beat.
// Snaps to the downbeats.
//
// Create a new m_normal and m_ramp and queues the ramp (RampQueued).
//
void TimeBase::insert_beat(void)
{
    HiJack::Client& client = getJackClientInstance();

    if(m_state == Ramping) {
        cout << "Already ramping... not skipping" << endl;
        return;
    }
    cout << "inserting beat..." << endl;
    jack_position_t pos;
    jack_transport_query(client.ref(), &pos);
    assert(pos.valid == JackPositionBBT);

    Epoch ramp, normal;

    ramp.m_beats_per_bar = (int32_t)pos.beats_per_bar;
    ramp.m_ticks_per_beat = (int32_t)pos.ticks_per_beat;
    if(pos.beat == ramp.m_beats_per_bar) {
        ramp.m_bar = pos.bar + 1;
        ramp.m_beat = 1;
    } else {
        ramp.m_bar = pos.bar;
        ramp.m_beat = pos.beat + 1;
    }
    ramp.m_tick = 0;

    normal = ramp;
    if(normal.m_beat >= (normal.m_beats_per_bar-1)) {
        ++normal.m_bar;
        normal.m_beat = normal.m_beat + 2 - normal.m_beats_per_bar;
    } else {
        normal.m_beat += 2;
    }

    QMutexLocker lock(&m_state_mutex);
    ramp.m_frame = m_normal.getFrame(ramp.m_bar, ramp.m_beat, ramp.m_tick);
    normal.m_frame = m_normal.getFrame(normal.m_bar, normal.m_beat, normal.m_tick);
    ramp.m_beats_per_minute = 0.5 * pos.beats_per_minute;
    normal.m_beats_per_minute = pos.beats_per_minute;
    if(normal.m_beat == 1) {
        --normal.m_bar;
        normal.m_beat = normal.m_beats_per_bar;
    } else {
        --normal.m_beat;
    }
    normal.init_frame_rates(pos.frame_rate);
    ramp.init_frame_rates(pos.frame_rate);
    m_normal_queued = normal;
    m_ramp = ramp;
    m_state = RampQueued;

    // Ok, we set the ramp... now it's up to to
    // timebase() to do the Right Thing.
}

void TimeBase::re_zero(void)
{
    QMutexLocker lock(&m_state_mutex);
    if(m_state == Ramping || m_state == RampQueued) {
        m_state = Steady;
        m_normal = m_normal_queued;
    }
    QWriteLocker e_mut(&m_normal.m_mutex);
    m_state = Steady;
    m_normal.m_bar = 1;
    m_normal.m_beat = 1;
    m_normal.m_tick = 0;
    m_normal.m_frame = 0;
    m_time_reset = true;
    e_mut.unlock();
    lock.unlock();
    jack_transport_locate(getJackClientInstance().ref(), 0);
}

void TimeBase::play(void)
{
    jack_transport_start(getJackClientInstance().ref());
}

void TimeBase::stop(void)
{
    jack_transport_stop(getJackClientInstance().ref());
}

unsigned TimeBase::getBeatsPerMeasure(void)
{
    QMutexLocker lock(&m_state_mutex);
    unsigned rv = 0;
    if(m_state == Ramping) {
        rv = m_ramp.m_beats_per_bar;
    } else {
        assert((m_state == RampQueued) || (m_state == Steady));
        rv = m_normal.m_beats_per_bar;
    }
    return rv;
}

// When rolling.  This is basically a ramp-and-snap-to,
// taking effect on the next available beat (possibly beat
// 1 of a new measure if we reduce the beats_per_bar.
// In fact, I copied the rampAndSnapTo() code as a starting point.
// When stopped, it does the same, except there's no ramping,
// and we don't increment the beat by one.
void TimeBase::setBeatsPerMeasure(unsigned ct)
{
    Epoch normal = m_normal;
    QMutexLocker lock(&m_state_mutex);
    jack_position_t pos;
    jack_transport_state_t trans_state = jack_transport_query(
        getJackClientInstance().ref(),
        &pos);
    jack_nframes_t curframe = pos.frame;

    // Note:  Looping is deprecated in the API.
    if ((trans_state == JackTransportStarting) || (trans_state == JackTransportLooping)) {
        return;
    }

//     if (trans_state == JackTransportStopped) {
//         if(normal.m_beat >= ct) {
//             normal.m_beat = 1;
//         }
//         normal.m_frames_per_measure *= ((double)ct)/((double)normal.m_beats_per_bar));
//         normal.m_beats_per_bar = ct;
//         return;
//     }

    assert((trans_state == JackTransportRolling) || (trans_state == JackTransportStopped));
    bool rolling = (trans_state == JackTransportRolling);

    // Find the nearest BBT for our new normal anchor
    jack_nframes_t ignored;
    normal.getBBT(curframe, &normal.m_bar, &normal.m_beat, &normal.m_tick, &ignored);
    if (rolling) ++normal.m_beat; // next beat
    if(normal.m_beat > normal.m_beats_per_bar) {
        ++normal.m_bar;
        normal.m_beat = 1;
    }
    if (rolling) normal.m_tick = 0;
    if (rolling) {
        normal.m_frame = m_normal.getFrame(normal.m_bar, normal.m_beat, normal.m_tick);
        if(normal.m_beat > ct) {
            ++normal.m_bar;
            normal.m_beat = 1;
        }
    } else {
        normal.m_frame = curframe;
    }
    normal.m_beats_per_bar = ct;
    normal.init_frame_rates(jack_get_sample_rate(getJackClientInstance().ref()));

    cerr << "Old normal: " << m_normal << endl;
    cerr << "New normal: " << normal << endl;
    if (rolling) {
        cerr << "Ramp:       " << normal << endl;
        m_normal_queued = normal;
        m_ramp = m_normal;
        m_state = RampQueued;
    } else {
        m_normal = normal;
    }
}

double TimeBase::getTempo(void)
{
    QMutexLocker lock(&m_state_mutex);
    double rv = 0;
    if(m_state == Ramping) {
        rv = m_ramp.m_beats_per_minute;
    } else {
        assert((m_state == RampQueued) || (m_state == Steady));
        rv = m_normal.m_beats_per_minute;
    }
    return rv;
}

void TimeBase::setTempo(double /*bpm*/)
{
#warning "TimeBase::setTempo() is still not implemented."
    assert(false); // not implemented yet.
}

