/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Performances           *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef INCONCERT_TIMEBASE_HPP__
#define INCONCERT_TIMEBASE_HPP__

/* TimeBase.hpp
   Gabriel M. Beddingfield
   2008-01-29

   TimeBase - This is the Jack Timebase master.  It keeps track of
   where you are and makes the adjustments that you
   request.

   Epoch - This struct is only for use with the class TimeBase.
   It provides a reference point from which time is
   measured.  For example, it will know that frame
   562891 was bar:beat:tick = 10:2:0134, and the tempo
   there is 97.0 beats per minute.  It can calculate
   the frame of 11:1:0000 (or any bar:beat:tick) and
   and also the bar:beat:tick for any frame.

*/

#include <string>
#include <memory> // for auto_ptr<>
#include <QMutex>
#include <jack/transport.h>
#include <jack/types.h>

#include "TapTempo/TapTempoInterface.hpp"
#include "HiJack/TimebaseMaster.hpp"
#include "Epoch.hpp"

namespace InConcert
{
// class TimeBase
//
// This class acts as a Jack Transport Master for synchronizing
// the computer audio with a live ensemble during the perfomance.
// It features a Tap Tempo interface and an interface to insert
// or skip a beat.
//
class TimeBase : public HiJack::TimebaseMaster
{
public:
    // Steady = m_normal is the current time base.
    //
    // Ramping = m_ramp is the current time base, until we reach
    //           m_normal.m_frame.
    //
    // RampQueued = when m_ramp.m_frame is reached, switch over to
    //              Ramping.
    //
    typedef enum { Steady, Ramping, RampQueued } timebase_state;

private:
    timebase_state m_state;
    QMutex m_state_mutex;
    // m_normal_queued is to hold the new m_normal, to be
    // applied _after_ ramping.
    Epoch m_normal, m_ramp, m_normal_queued;
    typedef TapTempoInterface<jack_nframes_t, double> tap_tempo_t;
    std::auto_ptr<tap_tempo_t> m_taps;
    bool m_time_reset;  // Reset the timebase now.
    bool m_auto_start;
    std::string m_name;

public:
    TimeBase();
    virtual ~TimeBase();

    timebase_state getState(void) const { return m_state; }

    // Tap settings
    unsigned getTapsRequired(void) const { return m_taps->getTapsRequired(); }
    void setTapsRequired(unsigned ct) { m_taps->setTapsRequired(ct); }
    double getTapsTimeout(void) const { return m_taps->getTimeout(); }
#warning "TODO:  User is thinking seconds... but TapTempo is *not*"
    void setTapsTimeout(double t) {
        m_taps->setTimeout(tap_tempo_t::scalar_type(t)*44100);
    }

    // methods

    // Methods needed by user
    void tap(void);
    void tap(jack_nframes_t time);  // jack_frame_time(), not the transport frame;
    void skip_beat(void);
    void insert_beat(void);
    void re_zero(void);
    void play(void);
    void stop(void);
    void setAutoStart(bool state) {
        m_auto_start = state;
    }
    bool getAutoStart(void) { return m_auto_start; }
    unsigned getBeatsPerMeasure(void);
    void setBeatsPerMeasure(unsigned ct);
    double getTempo(void);
    void setTempo(double bpm);

    // Jack timebase callback, routes to the private timebase_normal()
    // and timebase_ramping methods.
    const std::string& getName(void) { return m_name; }
    void timebaseCallback(
        jack_transport_state_t state,
        jack_nframes_t nframes,
        jack_position_t *pos,
        int new_pos);


private:
    // Jack timebase callback when ramping
    void timebase_ramping(
        jack_transport_state_t state,
        jack_nframes_t nframes,
        jack_position_t *pos,
        int new_pos);

    // Jack timebase callback when in normal (Steady) mode.
    void timebase_normal(
        jack_transport_state_t state,
        jack_nframes_t nframes,
        jack_position_t *pos,
        int new_pos);

    // Called after all the taps have been given to m_taps.
    //
    // Mutex must already be locked before calling this
    void rampAndSnapTo(jack_nframes_t target, double new_tempo);
}; // class TimeBase

} // namespace InConcert

#endif // INCONCERT_TIMEBASE_HPP__
